{
==============================================================================
 MapTextfiles - Release 99/1  14.03.1999
 for Delphi 4, should also run with Delphi 3

------------------------------------------------------------------------------

 MapTextfiles is FREEWARE. Freeware means, that you can use this software 
 without paying anything for it, but it is not public domain! This software 
 is protected through the law of the Federal Republic of Germany, the European
 Union and other countries.

 (C)1999 by Peter Hellinger Software, All rights are reserved.

 Peter Hellinger Software, Zerzabelshofstrasse 41, D-90480 Nrnberg

 email: mail@helli.de
 homepage: http://www.helli.de

==============================================================================

 Installation:
 Copy MAPTEXTFILE.PAS into your library path. Use ist. 8-)

------------------------------------------------------------------------------

 In Delphi you have many ways to manipulate text files. But all of these have 
 a handicap: You must read the whole file (StringList) or you can only access 
 data sequential (Assign/Readln).
        
 The Windows API provides the so called Memory Mapped Files. Windows self 
 uses MMFs to load EXE or DLL. Therefore the mechanism is very efficient, but
 simple to handle. The only handicap is, that you must know the size of the 
 file before you access it. This means for typical text files, that the 
 operation is normally limited to read from the file or manipulate inside.
        
 The class TMapTextfile wraps the neccesary API calls and povides efficent 
 functions for accessing the data.

 TMapTextfiles provides the following properties and methods:

 Methodes:
 =========

 Create         Creates an instace of the class

 Destroy        Destroys the instance

 Open           Opens a file as a memory mapped file.
 
                filename = name of the file
                mode =     open mode: 
                           mmRead =      Open only for read
                           mmReadWrite = Open for read and wrie 
                           
                Returns INVALID_HANDLE_VALUE if the file not exist.
                If the result > 0 all is OK.
                
                NOTE: 
                1. The file must exist!
                2. You cannot write behind the end of the file.
                
 Close          Close the memory mapped file and frees all handles.

 EndOfFile      Returns TRUE, if the End of the file is reached.
 
 GetSize        Returns the Size of the file in Bytes

 GetPos         Returns the actual file read/write position
                NOTE: Position 0 is the start of the file

 SetPos         Sets the actual read/write position
                NOTE: Position 0 is the start of the file

 ReadChar       Reads a character from the actual read/write position.
                The r/w position is incremented.

 ReadString     Returns a string, starting at the actual r/w position.
                The string is delimited by characters < SPACE, but not
                by ESC (#27) and TAB. The r/w position moves to the
                end of the string, delimiter chararcters are skiped.
 
 ReadLn         Same as ReadSring, but as a Procedure.
 
 ReadCharAt     Reads a charater from an arbitray possition.
                The r/w position is NOT moved!
          
                pos = position to read from (0 = start of the file!)
 
 ReadChars      Reads a line of charactes from the MMF.
                The r/w position is NOT moved!
 
                str = The buffer to read to
                pos = position to read from (0 = Start of the file!)
                len = number of characters to read.
 
 ReadStringAt   Returns a string, starting at an arbitray possition.
                The string is delimited by characters < SPACE, but not
                by ESC (#27) and TAB. The r/w position is NOT moved.

 FindString     Findes a substring in the MMF and Returns the position.
                
                str = rhe substring to search for
                pos = position to start the search (0 = start of the file)
                max = position to end the search. If this is 0 or less 
                      then 0 the end of the file is the limit.
                      
                Returns the position of the substring or -1 if the 
                substring is not found.
 
 FindWildCard   Same as Findstring, but supports wildcard search.
 
                str =   the substring to search for
                pos =   position to start the search (0 = start of the file)
                max =   position to end the search. If this is 0 or less 
                        then 0 the end of the file is the limit.
                wild =  the character used as wildcard (i.e. "*")
                joker = the character used as joker (i.e. "?")

                Returns the position of the substring or -1 if the 
                substring is not found.

 ReadBytes      Reads a number of bytes to a anonymous variable.
                The r/w position is NOT moved!
 
                b =   the anonymous variable
                pos = position to read from (0 = start of the file)
                len = number of bytes to read.
 
 WriteBytes     Writes a number of bytes to the file. 
                NOTE: You can not write behind the end of the file!!!
                
                b =   the anonymous variable
                pos = position to write to (0 = start of the file)
                len = number of bytes to write

==============================================================================
}

UNIT MapTextfile;

INTERFACE

USES Classes, Windows;

TYPE tMapMode =     (mmRead, mmReadWrite);

TYPE TMapTextfile = CLASS
                     PRIVATE
                      f_file:  THandle;
                      f_MMF:   THandle;
                      f_size:  INTEGER;
                      f_view:  PByte;
                      f_data:  PChar;
                      f_pos:   INTEGER;
                      f_open:  BOOLEAN;
                      FUNCTION CalcPos (pos: INTEGER): PCHAR;
                     PUBLIC
                      CONSTRUCTOR Create;
                      DESTRUCTOR Destroy; OVERRIDE;
                      FUNCTION  Open (CONST filename: STRING; mode: tMapMode): INTEGER;
                      PROCEDURE Close;
                      FUNCTION  ReadChar: CHAR;
                      FUNCTION  ReadString: STRING;
                      PROCEDURE ReadLn (VAR str: STRING);
                      FUNCTION  ReadCharAt (pos: LONGINT): CHAR;
                      PROCEDURE ReadChars (str: PCHAR; pos, len: LONGINT);
                      FUNCTION  ReadStringAt (pos: LONGINT): STRING;
                      FUNCTION  GetSize: LONGINT;
                      FUNCTION  GetPos:  LONGINT;
                      PROCEDURE SetPos (pos: LONGINT);
                      FUNCTION  EndOfFile: BOOLEAN;
                      FUNCTION  FindString (CONST str: STRING; pos, max: INTEGER): INTEGER;
                      FUNCTION  FindWildCard (CONST str: STRING; pos, max: INTEGER; wild, joker: CHAR): INTEGER;
                      PROCEDURE ReadBytes (VAR b; pos, len: LONGINT);
                      PROCEDURE WriteBytes (VAR b; pos, len: LONGINT);
                     END;

IMPLEMENTATION

CONSTRUCTOR TMapTextfile.Create;
BEGIN
 f_open:= FALSE;
END;


FUNCTION TMapTextfile.Open (CONST filename: STRING; mode: tMapmode): INTEGER;
VAR m1, m2, m3: CARDINAL;
BEGIN
 f_open:= FALSE;
 IF mode = mmRead THEN BEGIN
  m1:= GENERIC_READ;
  m2:= PAGE_READONLY;
  m3:= FILE_MAP_READ;
 END ELSE BEGIN
  m1:= GENERIC_READ + GENERIC_WRITE;
  m2:= PAGE_READWRITE;
  m3:= FILE_MAP_WRITE;
 END;

 f_file:= CreateFile (PCHAR (FileName), m1, FILE_SHARE_READ, NIL,
                      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
 // Sicherheitsprfung - alles ok ?
 IF f_file = INVALID_HANDLE_VALUE THEN BEGIN
  Result:= INVALID_HANDLE_VALUE;
  EXIT;
 END;
 TRY
  // wie gro ist der Brocken ?
  f_size:= GetFileSize (f_file, NIL);
  // Memory-Mapped-File-Objekt erzeugen
  f_MMF:= CreateFileMapping (f_file, NIL, m2, 0, f_size, NIL);
  // Sicherheitsprfung - immer noch alles ok ?
  IF f_MMF = 0 THEN BEGIN
   Result:= -1;
   EXIT;
  END;
 FINALLY
  // das Datei-Handle wird nun nicht mehr bentigt !
  CloseHandle (f_file);
 END;
 // Status: Datei-Handle freigegeben, MMF-Handle aktiv
 TRY
  // eine Sicht auf das MMF-Objekt einrichten
  f_View:= MapViewOfFile (f_MMF, m3, 0, 0, f_size);
  // Sicherheitsprfung - alles ok ?
  IF f_view = NIL THEN BEGIN
   Result:= -1;
   EXIT;
  END;
 FINALLY
  // Handle auf das MMF-Objekt wird nicht mehr bentigt
  CloseHandle (f_MMF);
 END;
 f_data:= PCHAR (f_view);
 f_pos:= 0;
 f_open:= TRUE;
 Result:= 0;
 // Status: Datei-Handle freigeben, MMF-Handle freigegeben, View ist aktiv
END;


DESTRUCTOR TMapTextfile.Destroy;
BEGIN
 IF f_open THEN Close;
 INHERITED;
END;


PROCEDURE TMapTextfile.Close;
BEGIN
 IF f_open THEN BEGIN
  UnmapViewOfFile (f_View);
  f_open:= FALSE;
 END;
END;

FUNCTION TMapTextfile.CalcPos (pos: INTEGER): PCHAR;
BEGIN
 Result:= NIL;
 IF f_open THEN BEGIN
  IF pos < 0 THEN pos:= 0;
  IF pos > f_size THEN pos:= f_size;
  result:= PCHAR (LONGINT(f_view) + pos);
 END;
END;

FUNCTION TMapTextfile.ReadChar: CHAR;
BEGIN
 Result:= #0;
 IF f_open THEN BEGIN
  f_data:= PCHAR (LONGINT(f_view) + f_pos);
  Result:= f_data^;
  INC (f_pos);
 END;
END;

FUNCTION TMapTextfile.ReadString: STRING;
BEGIN
 Result:= '';
 IF f_open THEN BEGIN
  f_data:= PCHAR (LONGINT(f_view) + f_pos);
  WHILE (f_pos < f_size) DO BEGIN
   CASE f_data^ OF
    #0..#31: CASE f_data^ OF
               #9,
              #27: Result:= Result + f_data^; // Tab und Escape weiterreichen
              #10: BEGIN  INC (f_pos);  EXIT;  END;
              #13: BEGIN // Carriage Return terminiert
                    INC (f_pos);
                    INC (f_data);
                    IF f_data^ = #10 THEN  INC (f_pos);
                    EXIT;
                   END;
             END;
    ELSE     Result:= Result + f_data^;
   END;
   INC (f_pos);
   INC (f_data);
  END;
 END;
END;

FUNCTION TMapTextfile.ReadCharAt (pos: LONGINT): CHAR;
BEGIN
 IF f_open THEN Result:= CalcPos (pos)^
           ELSE Result:= #0;
END;

PROCEDURE TMapTextfile.ReadChars (str: PCHAR; pos, len: LONGINT);
VAR i: INTEGER;
    p: PCHAR;
BEGIN
 IF f_open THEN BEGIN
  IF len <= 0 THEN EXIT;
  i:= 0;
  p:= CalcPos (pos);
  WHILE (i <= f_size) AND (i <= len) DO BEGIN
   str^:= p^;
   INC (str);
   INC (p);
   INC (i);
  END;
 END;
END;

PROCEDURE TMapTextfile.ReadBytes (VAR b; pos, len: LONGINT);
VAR p: PCHAR;
BEGIN
 IF f_open THEN BEGIN
  p:= CalcPos (pos);
  Move (p^, b, len);
 END;
END;

PROCEDURE TMapTextfile.WriteBytes (VAR b; pos, len: LONGINT);
VAR p: PCHAR;
BEGIN
 IF f_open THEN BEGIN
  p:= CalcPos (pos);
  Move (b, p^, len);
 END;
END;


FUNCTION TMapTextfile.ReadStringAt (pos: LONGINT): STRING;
VAR i: INTEGER;
    p: PCHAR;
BEGIN
 Result:= '';
 IF f_open THEN BEGIN
  p:= CalcPos (pos);
  i:= 0;
  WHILE (i <= f_size) DO BEGIN
   CASE p^ OF
    #0..#31: CASE p^ OF
               #9,
              #27: Result:= Result + p^; // Tabs und Escape weiterreichen
              #10,
              #13: EXIT; // Linefeed and Carriage Return terminiert
             END;
    ELSE  Result:= Result + p^;
   END;
   INC (p);
  END;
 END;
END;

PROCEDURE TMapTextfile.ReadLn (VAR str: STRING);
BEGIN
 str:= ReadString;
END;

FUNCTION TMapTextfile.GetSize: LONGINT;
BEGIN
 IF f_open THEN Result:= f_size
           ELSE Result:= -1;
END;

FUNCTION TMapTextfile.GetPos:  LONGINT;
BEGIN
 IF f_open THEN Result:= f_pos
           ELSE Result:= -1;
END;

PROCEDURE TMapTextfile.SetPos (pos: LONGINT);
BEGIN
 IF f_open THEN BEGIN
  IF pos < 0 THEN pos:= 0;
  IF pos > f_size THEN pos:= f_size;
  f_pos:= pos;
 END;
END;

FUNCTION TMapTextfile.EndOfFile: BOOLEAN;
BEGIN
 IF f_open THEN Result:= f_pos >= f_size
           ELSE Result:= TRUE;
END;

FUNCTION TMapTextfile.FindString (CONST str: STRING; pos, max: INTEGER): INTEGER;
VAR s, l1, j: INTEGER;
    p, x: PCHAR;
BEGIN
 Result:= -1;
 IF f_open THEN BEGIN
  IF max <= 0 THEN  max:= f_size;
  IF pos < 0 THEN pos:= f_pos;
  IF pos > max THEN EXIT;
  x:= PCHAR (str);
  p:= PCHAR (f_view);
  l1:= 0;  WHILE (x[l1] > #0) DO INC (l1);
  IF (l1 > 0) THEN BEGIN
   s:= pos;
   REPEAT (* 1 *)
    j:= 0;
    WHILE (s+j < max)
      AND (j < l1)
      AND (x[j] = p[s+j]) DO BEGIN
     INC (j);
     IF (j = l1) THEN BEGIN  Result:= s;  EXIT;  END;
    END;
    INC (s);
   UNTIL s >= f_size;
  END;
 END;
END;

FUNCTION TMapTextfile.FindWildCard (CONST str: STRING; pos, max: INTEGER;
                                    wild, joker: CHAR): INTEGER;
VAR s, l1, j: INTEGER;
    p, x: PCHAR;
BEGIN
 Result:= -1;
 IF f_open THEN BEGIN
  IF max <= 0 THEN  max:= f_size;
  IF pos < 0 THEN pos:= f_pos;
  IF pos > max THEN EXIT;

   x:= PCHAR (str);
   p:= PCHAR (f_view);
   l1:= 0;  WHILE (x[l1] > #0) DO INC (l1);

  IF (l1 > 0) THEN BEGIN
   s:= pos;
   REPEAT (* 1 *)
    j:= 0;
    WHILE (s+j < max)
      AND (j < l1)
      AND ((x[j] = p[s+j]) OR (x[j] = joker)) DO BEGIN
     INC (j);
     IF (x[j] = wild) OR (j >= l1) THEN BEGIN
      Result:= s;  EXIT;
     END;
    END;
    INC (s);
   UNTIL s >= f_size;
  END;
 END;
END;


END.
